##############################################################################
#
# Copyright (c) 2006 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Mail delivery names vocabulary test
"""

import doctest
import unittest

from zope.component.testing import PlacelessSetup


class MailDeliveryNamesTests(PlacelessSetup, unittest.TestCase):

    _marker = object()

    def _register(self, *names):
        from zope.component import provideUtility
        from zope.interface import implementer
        from zope.sendmail.interfaces import IMailDelivery
        from zope.sendmail.vocabulary import MailDeliveryNames

        @implementer(IMailDelivery)
        class StubMailDelivery(object):
            pass

        for name in names:
            provideUtility(StubMailDelivery(), name=name)

        provideUtility(MailDeliveryNames, name='Mail Delivery Names')

    def _callFUT(self):
        from zope.sendmail.vocabulary import MailDeliveryNames
        return MailDeliveryNames()

    def test_default(self):
        NAMES = 'and now for something completely different'.split()
        self._register(*NAMES)
        vocab = self._callFUT()
        names = sorted([term.value for term in vocab])
        self.assertEqual(names, sorted(NAMES))


def test_suite():
    return unittest.TestSuite([
        unittest.defaultTestLoader.loadTestsFromName(__name__),
        doctest.DocTestSuite('zope.sendmail.vocabulary'),
    ])



##############################################################################
#
# Copyright (c) 2003 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Queue processor thread

This module contains the queue processor thread.
"""
__docformat__ = 'restructuredtext'

import atexit
import logging
import argparse
import os
import smtplib
import threading
import time
import errno

from zope.sendmail.maildir import Maildir
from zope.sendmail.mailer import SMTPMailer

import sys
if sys.platform == 'win32':  # pragma: no cover
    import win32file
    import winerror
    import pywintypes

    def _os_link(src, dst):
        return win32file.CreateHardLink(dst, src, None)
else:
    _os_link = os.link
    pywintypes = None

try:
    import ConfigParser as configparser
except ImportError:  # pragma: PY3
    import configparser

# The longest time sending a file is expected to take.  Longer than this and
# the send attempt will be assumed to have failed.  This means that sending
# very large files or using very slow mail servers could result in duplicate
# messages sent.
MAX_SEND_TIME = 60*60*3

# The below diagram depicts the operations performed while sending a message in
# the ``run`` method of ``QueueProcessorThread``.  This sequence of operations
# will be performed for each file in the maildir each time the thread "wakes
# up" to send messages.
#
# Any error conditions not depicted on the diagram will provoke the catch-all
# exception logging of the ``run`` method.
#
# In the diagram the "message file" is the file in the maildir's "cur"
# directory that contains the message and "tmp file" is a hard link to the
# message file created in the maildir's "tmp" directory.
#
#           ( start trying to deliver a message )
#                            |
#                            |
#                            V
#            +-----( get tmp file mtime )
#            |               |
#            |               | file exists
#            |               V
#            |         ( check age )-----------------------------+
#   tmp file |               |                       file is new |
#   does not |               | file is old                       |
#   exist    |               |                                   |
#            |      ( unlink tmp file )-----------------------+  |
#            |               |                      file does |  |
#            |               | file unlinked        not exist |  |
#            |               V                                |  |
#            +---->( touch message file )------------------+  |  |
#                            |                   file does |  |  |
#                            |                   not exist |  |  |
#                            V                             |  |  |
#            ( link message file to tmp file )----------+  |  |  |
#                            |                 tmp file |  |  |  |
#                            |           already exists |  |  |  |
#                            |                          |  |  |  |
#                            V                          V  V  V  V
#                     ( send message )             ( skip this message )
#                            |
#                            V
#                 ( unlink message file )---------+
#                            |                    |
#                            | file unlinked      | file no longer exists
#                            |                    |
#                            |  +-----------------+
#                            |  |
#                            |  V
#                  ( unlink tmp file )------------+
#                            |                    |
#                            | file unlinked      | file no longer exists
#                            V                    |
#                  ( message delivered )<---------+


class QueueProcessorThread(threading.Thread):
    """This thread is started at configuration time from the
    `mail:queuedDelivery` directive handler if processorThread is True.
    """

    log = logging.getLogger("QueueProcessorThread")
    _stopped = False
    interval = 3.0   # process queue every X second
    maildir = None
    mailer = None

    def __init__(self, interval=3.0):
        threading.Thread.__init__(
            self, name="zope.sendmail.queue.QueueProcessorThread")
        self.interval = interval
        self._lock = threading.Lock()
        self.setDaemon(True)

    def setMaildir(self, maildir):
        """Set the maildir.

        This method is used just to provide a `maildir` stub.
        """
        self.maildir = maildir

    def _makeMaildir(self, path):
        return Maildir(path, True)

    def setQueuePath(self, path):
        self.setMaildir(self._makeMaildir(path))

    def setMailer(self, mailer):
        self.mailer = mailer

    def _parseMessage(self, message):
        """Extract fromaddr and toaddrs from the first two lines of
        the `message`.

        Returns a fromaddr string, a toaddrs tuple and the message
        string.
        """

        fromaddr = ""
        toaddrs = ()
        rest = ""

        try:
            first, second, rest = message.split(b'\n', 2)
        except ValueError:
            return fromaddr, toaddrs, message

        if first.startswith(b"X-Zope-From: "):
            i = len(b"X-Zope-From: ")
            fromaddr = first[i:].decode()

        if second.startswith(b"X-Zope-To: "):
            i = len(b"X-Zope-To: ")
            toaddrs = tuple(
                addr.decode() for addr in second[i:].split(b", ") if addr
            )

        return fromaddr, toaddrs, rest

    def _action_if_exists(self, fname, func, default=None):
        # apply the func to the fname, ignoring exceptions that
        # happen when the file does not exist.
        try:
            return func(fname)
        except OSError as e:
            if e.errno != errno.ENOENT:
                # The file existed, but something unexpected
                # happened. Report it
                raise
            return default

    def _unlink_if_exists(self, fname):
        self._action_if_exists(fname, os.unlink)

    def run(self, forever=True):
        atexit.register(self.stop)
        while not self._stopped:
            for filename in self.maildir:
                # if we are asked to stop while sending messages, do so
                if self._stopped:
                    break
                self._process_one_file(filename)
            else:
                if forever:
                    time.sleep(self.interval)

            # A testing plug
            if not forever:
                break

    def _process_one_file(self, filename):
        fromaddr = ''
        toaddrs = ()
        head, tail = os.path.split(filename)
        tmp_filename = os.path.join(head, '.sending-' + tail)
        rejected_filename = os.path.join(head, '.rejected-' + tail)
        try:
            # perform a series of operations in an attempt to ensure
            # that no two threads/processes send this message
            # simultaneously as well as attempting to not generate
            # spurious failure messages in the log; a diagram that
            # represents these operations is included in a
            # comment above this class

            # find the age of the tmp file (if it exists)
            mtime = self._action_if_exists(
                tmp_filename,
                lambda fname: os.stat(fname).st_mtime)
            age = time.time() - mtime if mtime is not None else None

            # if the tmp file exists, check its age
            if age is not None:
                if age > MAX_SEND_TIME:
                    # the tmp file is "too old"; this suggests
                    # that during an attempt to send it, the
                    # process died; remove the tmp file so we
                    # can try again
                    try:
                        os.unlink(tmp_filename)
                    except OSError as e:
                        if e.errno == errno.ENOENT:  # file does not exist
                            # it looks like someone else removed the tmp
                            # file, that's fine, we'll try to deliver the
                            # message again later
                            return
                        # XXX: we're silently ignoring the exception here.
                        # Is that right?
                        # If permissions or something are not right, we'll fail
                        # on _os_link later on.
                    # if we get here, the file existed, but was too
                    # old, so it was unlinked
                else:
                    # the tmp file is "new", so someone else may
                    # be sending this message, try again later
                    return

            # now we know (hope, given the above XXX) that the
            # tmp file doesn't exist, we need to "touch" the
            # message before we create the tmp file so the
            # mtime will reflect the fact that the file is
            # being processed (there is a race here, but it's
            # OK for two or more processes to touch the file
            # "simultaneously")
            try:
                os.utime(filename, None)
            except OSError as e:
                if e.errno == errno.ENOENT:  # file does not exist
                    # someone removed the message before we could
                    # touch it, no need to complain, we'll just keep
                    # going
                    return
                # XXX: Silently ignoring all other errors

            # creating this hard link will fail if another process is
            # also sending this message
            try:
                _os_link(filename, tmp_filename)
            except OSError as e:
                if e.errno == errno.EEXIST:  # file exists, *nix
                    # it looks like someone else is sending this
                    # message too; we'll try again later
                    return
                # XXX: Silently ignoring all other errno
            except Exception as e:  # pragma: no cover
                if (pywintypes is not None
                        and isinstance(e, pywintypes.error)
                        and e.funcname == 'CreateHardLink'
                        and e.winerror == winerror.ERROR_ALREADY_EXISTS):
                    # file exists, win32
                    return
                # XXX: Silently ignoring all other causes here.

            # read message file and send contents
            with open(filename, 'rb') as f:
                message = f.read()

            fromaddr, toaddrs, message = self._parseMessage(message)
            # The next block is the only one that is sensitive to
            # interruptions.  Everywhere else, if this daemon thread
            # stops, we should be able to correctly handle a restart.
            # In this block, if we send the message, but we are
            # stopped before we unlink the file, we will resend the
            # message when we are restarted.  We limit the likelihood
            # of this somewhat by using a lock to link the two
            # operations.  When the process gets an interrupt, it
            # will call the atexit that we registered (``stop``
            # below).  This will try to get the same lock before it
            # lets go.  Because this can cause the daemon thread to
            # continue (that is, to not act like a daemon thread), we
            # still use the _stopped flag to communicate.
            with self._lock:
                try:
                    self.mailer.send(fromaddr, toaddrs, message)
                except smtplib.SMTPResponseException as e:
                    if 500 <= e.smtp_code <= 599:
                        # permanent error, ditch the message
                        self.log.error(
                            "Discarding email from %s to %s due to"
                            " a permanent error: %s",
                            fromaddr, ", ".join(toaddrs), str(e))
                        _os_link(filename, rejected_filename)
                    else:
                        # Log an error and retry later
                        raise
                except smtplib.SMTPRecipientsRefused as e:
                    # All recipients are refused by smtp
                    # server. Dont try to redeliver the message.
                    self.log.error("Email recipients refused: %s",
                                   ', '.join(e.recipients))
                    _os_link(filename, rejected_filename)

                self._unlink_if_exists(filename)

            self._unlink_if_exists(tmp_filename)

            # TODO: maybe log the Message-Id of the message sent
            self.log.info("Mail from %s to %s sent.",
                          fromaddr, ", ".join(toaddrs))
            # Blanket except because we don't want
            # this thread to ever die
        except Exception:
            if fromaddr != '' or toaddrs != ():
                self.log.error(
                    "Error while sending mail from %s to %s.",
                    fromaddr, ", ".join(toaddrs), exc_info=True)
            else:
                self.log.error(
                    "Error while sending mail : %s ",
                    filename, exc_info=True)

    def stop(self):
        self._stopped = True
        self._lock.acquire()
        self._lock.release()


def boolean(s):
    s = str(s).lower()
    return s.startswith("t") or s.startswith("y") or s.startswith("1")


def string_or_none(s):
    if s == 'None':
        return None
    return s


class ConsoleApp(object):
    """Allows running of Queue Processor from the console."""

    INI_SECTION = "app:zope-sendmail"
    INI_NAMES = [
        "interval",
        "hostname",
        "port",
        "username",
        "password",
        "force_tls",
        "no_tls",
        "queue_path",
    ]

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--daemon', action='store_true',
        help=("Run in daemon mode, periodically checking queue "
              "and sending messages.  Default is to send all "
              "messages in queue once and exit."))
    parser.add_argument(
        '--interval', metavar='<#secs>', type=float, default=3,
        help=("How often to check queue when in daemon mode. "
              "Default is %(default)s seconds."))
    smtp_group = parser.add_argument_group(
        "SMTP Server",
        "Connection information for the SMTP server")
    smtp_group.add_argument(
        '--hostname', default='localhost',
        help=("Name of SMTP host to use for delivery.  Default is "
              "%(default)s."))
    smtp_group.add_argument(
        '--port', type=int, default=25,
        help=("Which port on SMTP server to deliver mail to. "
              "Default is %(default)s."))

    auth_group = parser.add_argument_group(
        "Authentication",
        ("Authentication information for the SMTP server. "
         "If one is provided, they must both be. One or both "
         "can be provided in the --config file."))
    auth_group.add_argument(
        '--username',
        help=("Username to use to log in to SMTP server.  Default "
              "is none."))
    auth_group.add_argument(
        '--password',
        help=("Password to use to log in to SMTP server.  Must be "
              "specified if username is specified."))
    del auth_group
    tls_group = smtp_group.add_mutually_exclusive_group()
    tls_group.add_argument(
        '--force-tls', action='store_true',
        help=("Do not connect if TLS is not available.  Not "
              "enabled by default."))
    tls_group.add_argument(
        '--no-tls', action='store_true',
        help=("Do not use TLS even if is available.  Not enabled "
              "by default."))
    del tls_group
    del smtp_group
    parser.add_argument(
        '--config', metavar='<inifile>',
        type=argparse.FileType(),
        help=("Get configuration from specified ini file; it must "
              "contain a section [%s] that can contain the "
              "following keys: %s. If you specify the queue path "
              "in the ini file, you don't need to specify it on "
              "the command line. With the exception of the queue path, "
              "options specified in the ini file override options on the "
              "command line." % (INI_SECTION, ', '.join(INI_NAMES))))
    parser.add_argument(
        "maildir", default=None, nargs="?",
        help=("The path to the mail queue directory."
              "If not given, it must be found in the --config file."
              "If given, this overrides a value in the --config file"))

    daemon = False
    interval = 3
    hostname = 'localhost'
    port = 25
    username = None
    password = None
    force_tls = False
    no_tls = False
    queue_path = None

    QueueProcessorKind = QueueProcessorThread
    MailerKind = SMTPMailer

    def __init__(self, argv=None, verbose=True):
        argv = sys.argv if argv is None else argv
        self.script_name = argv[0]
        self.verbose = verbose
        self._process_args(argv[1:])
        self.mailer = self.MailerKind(
            self.hostname, self.port, self.username, self.password,
            self.no_tls, self.force_tls)

    def main(self):
        queue = self.QueueProcessorKind(self.interval)
        queue.setMailer(self.mailer)
        queue.setQueuePath(self.queue_path)
        queue.run(forever=self.daemon)

    def _process_args(self, args):
        opts = self.parser.parse_args(args)
        self.daemon = opts.daemon
        self.interval = opts.interval
        self.hostname = opts.hostname
        self.port = opts.port
        self.username = opts.username
        self.password = opts.password
        self.force_tls = opts.force_tls
        self.no_tls = opts.no_tls

        if opts.config:
            fname = opts.config.name
            opts.config.close()
            self._load_config(fname)
        self.queue_path = opts.maildir or self.queue_path

        if not self.queue_path:
            self.parser.error('please specify the queue path')
        if (self.username or self.password) and \
           not (self.username and self.password):
            self.parser.error('Must use username and password together.')

    def _load_config(self, path):
        section = self.INI_SECTION
        names = self.INI_NAMES
        defaults = dict([(name, str(getattr(self, name))) for name in names])
        config = configparser.ConfigParser(defaults)
        config.read(path)
        self.interval = float(config.get(section, "interval"))
        self.hostname = config.get(section, "hostname")
        self.port = int(config.get(section, "port"))
        self.username = string_or_none(config.get(section, "username"))
        self.password = string_or_none(config.get(section, "password"))
        self.force_tls = boolean(config.get(section, "force_tls"))
        self.no_tls = boolean(config.get(section, "no_tls"))
        self.queue_path = string_or_none(config.get(section, "queue_path"))


def run(argv=None):
    logging.basicConfig()
    app = ConsoleApp(argv)
    app.main()


if __name__ == '__main__':
    run()



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Test the gts ZCML namespace directives.
"""
import os
import shutil
import unittest
import tempfile


import zope.component
from zope.component.testing import PlacelessSetup
from zope.configuration import xmlconfig
from zope.interface import implementer

from zope.sendmail.interfaces import \
     IMailDelivery, IMailer, ISMTPMailer
from zope.sendmail import delivery
from zope.sendmail import zcml
import zope.sendmail.tests


class MaildirStub(object):
    pass


@implementer(IMailer)
class Mailer(object):
    pass


class MockQueueProcessorThread(object):

    def setMailer(self, mailer):
        pass

    setQueuePath = setMailer

    def start(self):
        pass


class DirectivesTest(PlacelessSetup, unittest.TestCase):

    def setUp(self):
        super(DirectivesTest, self).setUp()

        self.mailbox = os.path.join(tempfile.mkdtemp(), "mailbox")
        self.addCleanup(shutil.rmtree, self.mailbox, True)
        self.testMailer = Mailer()
        self.smtpmailer = Mailer()
        gsm = zope.component.getGlobalSiteManager()

        gsm.registerUtility(self.smtpmailer, IMailer, "test.smtp")
        gsm.registerUtility(self.testMailer, IMailer, "test.mailer")

        here = os.path.dirname(__file__)
        with open(os.path.join(here, "mail.zcml"), 'r') as f:
            self.zcml = f.read()
        self.zcml = self.zcml.replace('path/to/tmp/mailbox', self.mailbox)

        self.orig_quethread = zcml.QueueProcessorThread
        self.orig_maildir = delivery.Maildir
        delivery.Maildir = MaildirStub
        zcml.QueueProcessorThread = MockQueueProcessorThread

        self.context = xmlconfig.string(self.zcml)

    def tearDown(self):
        delivery.Maildir = self.orig_maildir
        zcml.QueueProcessorThread = self.orig_quethread

        super(DirectivesTest, self).tearDown()

    def testQueuedDelivery(self):
        delivery = zope.component.getUtility(IMailDelivery, "Mail")
        self.assertEqual('QueuedMailDelivery', delivery.__class__.__name__)
        self.assertEqual(self.mailbox, delivery.queuePath)

    def testDirectDelivery(self):
        delivery = zope.component.getUtility(IMailDelivery, "Mail2")
        self.assertEqual('DirectMailDelivery', delivery.__class__.__name__)
        self.assertTrue(self.testMailer is delivery.mailer)

    def testSMTPMailer(self):
        mailer = zope.component.getUtility(IMailer, "smtp")
        self.assertTrue(ISMTPMailer.providedBy(mailer))

    def _check_zcml_without_registration(self, utility, name):
        gsm = zope.component.getGlobalSiteManager()
        gsm.unregisterUtility(utility, IMailer, name)
        from zope.configuration.exceptions import ConfigurationError
        with self.assertRaises(ConfigurationError) as exc:
            xmlconfig.string(self.zcml)

        msg = str(exc.exception)
        self.assertIn(name, msg)
        self.assertIn('is not defined', msg)

    def test_zcml_without_registered_smtp_mailer(self):
        self._check_zcml_without_registration(self.smtpmailer, 'test.smtp')

    def test_zcml_without_registered_mailer(self):
        self._check_zcml_without_registration(self.testMailer, 'test.mailer')



# coding=utf-8
##############################################################################
#
# Copyright (c) 2003 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Classes which abstract different channels a message could be sent to.
"""
__docformat__ = 'restructuredtext'

from threading import local
from ssl import SSLError
from smtplib import SMTP

from zope.interface import implementer
from zope.sendmail.interfaces import ISMTPMailer
from zope.sendmail._compat import PY2
from zope.sendmail._compat import text_type


class _SMTPState(local):
    connection = None
    code = None
    response = None


@implementer(ISMTPMailer)
class SMTPMailer(object):
    """Implementation of :class:`zope.sendmail.interfaces.ISMTPMailer`."""

    smtp = SMTP

    def __init__(self, hostname='localhost', port=25,
                 username=None, password=None, no_tls=False, force_tls=False):
        self.hostname = hostname
        self.port = port
        self.username = username
        self.password = password
        self.force_tls = force_tls
        self.no_tls = no_tls
        self._smtp = _SMTPState()

    def _make_property(name):
        return property(lambda self: getattr(self._smtp, name),
                        lambda self, nv: setattr(self._smtp, name, nv))

    connection = _make_property('connection')
    code = _make_property('code')
    response = _make_property('response')

    del _make_property

    def vote(self, fromaddr, toaddrs, message):
        self.connection = self.smtp(self.hostname, str(self.port))

        code, response = self.connection.ehlo()
        if code < 200 or code >= 300:
            code, response = self.connection.helo()
            if code < 200 or code >= 300:
                raise RuntimeError('Error sending HELO to the SMTP server '
                                   '(code=%s, response=%s)' % (code, response))

        self.code, self.response = code, response

    def _close_connection(self):
        try:
            self.connection.quit()
        except SSLError:
            # something weird happened while quiting
            self.connection.close()
        self.connection = None

    def abort(self):
        if self.connection is None:
            return
        self._close_connection()

    def send(self, fromaddr, toaddrs, message):
        connection = self.connection
        if connection is None:
            self.vote(fromaddr, toaddrs, message)

        connection = self.connection

        # encryption support
        have_tls = connection.has_extn('starttls')
        if not have_tls and self.force_tls:
            raise RuntimeError('TLS is not available but TLS is required')

        if have_tls and not self.no_tls:
            connection.starttls()
            connection.ehlo()

        if connection.does_esmtp:
            if self.username is not None and self.password is not None:
                username, password = self.username, self.password
                if PY2 and isinstance(username, text_type):
                    username = username.encode('utf-8')  # pragma: PY2
                if PY2 and isinstance(password, text_type):
                    password = password.encode('utf-8')  # pragma: PY2
                connection.login(username, password)
        elif self.username:
            raise RuntimeError(
                'Mailhost does not support ESMTP but a username is configured')

        try:
            connection.sendmail(fromaddr, toaddrs, message)
        finally:
            self._close_connection()



##############################################################################
#
# Copyright (c) 2003 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Mail Delivery Tests

Simple implementation of the MailDelivery, Mailers and MailEvents.
"""
import sys
import smtplib
import unittest
import os.path
import tempfile
import shutil

import transaction
from zope.interface import implementer
from zope.interface.verify import verifyObject

from zope.sendmail.interfaces import IMailer
from zope.sendmail.interfaces import IDirectMailDelivery
from zope.sendmail.delivery import AbstractMailDelivery
from zope.sendmail.delivery import DirectMailDelivery


@implementer(IMailer)
class MailerStub(object):

    def __init__(self, *args, **kw):
        self.sent_messages = []

    def send(self, fromaddr, toaddrs, message):
        self.sent_messages.append((fromaddr, toaddrs, message))

    abort = None
    vote = None


class TestMailDataManager(unittest.TestCase):

    def testInterface(self):
        from transaction.interfaces import IDataManager
        from zope.sendmail.delivery import MailDataManager
        manager = MailDataManager(object, (1, 2))
        verifyObject(IDataManager, manager)
        self.assertEqual(manager.callable, object)
        self.assertEqual(manager.args, (1, 2))
        # required by IDataManager
        self.assertIsInstance(manager.sortKey(), str)

    def test_successful_commit(self):
        # Regression test for http://www.zope.org/Collectors/Zope3-dev/590
        from zope.sendmail.delivery import MailDataManager

        _success = []

        def _on_success(*args):
            _success.append(args)

        def _on_abort(*args):
            self.fail("Should not abort")

        manager = MailDataManager(_on_success, ('foo', 'bar'),
                                  onAbort=_on_abort)
        xact = object()
        manager.tpc_begin(xact)
        manager.commit(xact)
        manager.tpc_vote(xact)
        manager.tpc_finish(xact)
        self.assertEqual(_success, [('foo', 'bar')])

    def test_unsuccessful_commit(self):
        # Regression test for http://www.zope.org/Collectors/Zope3-dev/590
        from zope.sendmail.delivery import MailDataManager

        _success = []
        _abort = []

        def _on_success(*args):
            self.fail("Should not succeed")

        def _on_abort(*args):
            _abort.append(args)
        manager = MailDataManager(_on_success, ('foo', 'bar'),
                                  onAbort=_on_abort)
        xact = object()
        manager.tpc_begin(xact)
        manager.commit(xact)
        manager.tpc_vote(xact)
        manager.tpc_abort(xact)
        self.assertEqual(_success, [])


class TestAbstractMailDelivery(unittest.TestCase):

    # Avoid DeprecationWarning for assertRaisesRegexp on Python 3 while
    # coping with Python 2 not having the Regex spelling variant
    assertRaisesRegex = getattr(unittest.TestCase, 'assertRaisesRegex',
                                unittest.TestCase.assertRaisesRegexp)

    def test_bad_message_id(self):
        class Parser(object):
            def parsestr(self, s, headersonly=False):
                return {'Message-Id': 'bad id'}

        import email.parser
        orig_parser = email.parser.Parser
        self.addCleanup(setattr, email.parser, 'Parser', orig_parser)
        email.parser.Parser = Parser

        delivery = AbstractMailDelivery()
        with self.assertRaisesRegex(ValueError,
                                    "Malformed Message-Id header"):
            delivery.send(None, None, None)


class TestDirectMailDelivery(unittest.TestCase):

    def testInterface(self):
        mailer = MailerStub()
        delivery = DirectMailDelivery(mailer)
        verifyObject(IDirectMailDelivery, delivery)
        self.assertEqual(delivery.mailer, mailer)

    def testSend(self, send_unicode=False, message=None):
        mailer = MailerStub()
        delivery = DirectMailDelivery(mailer)
        fromaddr = 'Jim <jim@example.com'
        toaddrs = ('Guido <guido@example.com>',
                   'Steve <steve@examplecom>')
        opt_headers = (b'From: Jim <jim@example.org>\n'
                       b'To: some-zope-coders:;\n'
                       b'Date: Mon, 19 May 2003 10:17:36 -0400\n'
                       b'Message-Id: <20030519.1234@example.org>\n')
        if message is None:
            message = (b'Subject: example\n'
                       b'\n'
                       b'This is just an example\n')

        if send_unicode:
            opt_headers_bytes = opt_headers
            opt_headers = opt_headers.decode('utf-8')
            message_bytes = message + b'\xc3\xa4'
            message = message_bytes.decode('utf-8')
        else:
            message_bytes = message
            opt_headers_bytes = opt_headers

        msgid = delivery.send(fromaddr, toaddrs, opt_headers + message)
        self.assertEqual(msgid, '20030519.1234@example.org')
        self.assertEqual(mailer.sent_messages, [])
        transaction.commit()
        self.assertEqual(mailer.sent_messages,
                         [(fromaddr, toaddrs,
                           opt_headers_bytes + message_bytes)])

        mailer.sent_messages = []
        msgid = delivery.send(fromaddr, toaddrs, message)
        self.assertIn('@', msgid)
        self.assertEqual(mailer.sent_messages, [])
        transaction.commit()
        self.assertEqual(len(mailer.sent_messages), 1)
        self.assertEqual(mailer.sent_messages[0][0], fromaddr)
        self.assertEqual(mailer.sent_messages[0][1], toaddrs)
        self.assertTrue(mailer.sent_messages[0][2].endswith(message_bytes))
        new_headers = mailer.sent_messages[0][2][:-len(message_bytes)]
        self.assertIn(('Message-Id: <%s>' % msgid).encode(), new_headers)

        mailer.sent_messages = []
        msgid = delivery.send(fromaddr, toaddrs, opt_headers + message)
        self.assertEqual(mailer.sent_messages, [])
        transaction.abort()
        self.assertEqual(mailer.sent_messages, [])

    def testSendUnicode(self):
        self.testSend(send_unicode=True)

    def testSendLatin1(self):
        '''
        Test to send a mail that is not valid UTF-8. Since we are using bytes
        everywhere, this is not a problem.
        '''
        message = (b'Subject: example\n'
                   b'Content-Type: text/plain; charset="latin1"\n'
                   b'\n'
                   b'\xfc')
        self.testSend(message=message)

    def testBrokenMailerErrorsAreEaten(self):
        from zope.testing.loggingsupport import InstalledHandler
        mailer = BrokenMailerStub()
        delivery = DirectMailDelivery(mailer)
        fromaddr = 'Jim <jim@example.com'
        toaddrs = ('Guido <guido@example.com>',
                   'Steve <steve@examplecom>')
        opt_headers = ('From: Jim <jim@example.org>\n'
                       'To: some-zope-coders:;\n'
                       'Date: Mon, 19 May 2003 10:17:36 -0400\n'
                       'Message-Id: <20030519.1234@example.org>\n')
        message = ('Subject: example\n'
                   '\n'
                   'This is just an example\n')

        delivery.send(fromaddr, toaddrs, opt_headers + message)
        log_handler = InstalledHandler('MailDataManager')
        self.addCleanup(log_handler.uninstall)
        self.addCleanup(transaction.abort)
        transaction.commit()

    def testRefusingMailerDiesInVote(self):
        mailer = RefusingMailerStub()
        delivery = DirectMailDelivery(mailer)
        fromaddr = 'Jim <jim@example.com'
        toaddrs = ('Guido <guido@example.com>',
                   'Steve <steve@examplecom>')
        opt_headers = ('From: Jim <jim@example.org>\n'
                       'To: some-zope-coders:;\n'
                       'Date: Mon, 19 May 2003 10:17:36 -0400\n'
                       'Message-Id: <20030519.1234@example.org>\n')
        message = ('Subject: example\n'
                   '\n'
                   'This is just an example\n')

        delivery.send(fromaddr, toaddrs, opt_headers + message)
        self.addCleanup(transaction.abort)

        with self.assertRaises(Exception):
            transaction.commit()
        self.assertFalse(transaction.get()._voted,
                         "We voted for commit then failed, reraise")

    def test_old_mailer_without_vote(self):
        import warnings

        class OldMailer(object):
            def send(self):
                raise NotImplementedError()
            abort = send

        delivery = DirectMailDelivery(OldMailer())
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')

            mdm = delivery.createDataManager("from", (), "msg")
            mdm.tpc_vote(None)

        self.assertEqual(1, len(w))
        self.assertIn("does not provide a vote method", str(w[0]))

    def testSavepoint(self):
        mailer = MailerStub()
        delivery = DirectMailDelivery(mailer)
        fromaddr = 'Jim <jim@example.com'
        toaddrs = ('Guido <guido@example.com>',)
        delivery.send(fromaddr, toaddrs, b'Subject: one')
        # Reminder: nothing is sent yet, sending happens after commit.
        self.assertEqual(mailer.sent_messages, [])
        # We create a savepoint.  If we rollback to this savepoint,
        # the previous mail should still remain in the queue.
        savepoint = transaction.savepoint()
        delivery.send(fromaddr, toaddrs, b'Subject: two')
        self.assertEqual(mailer.sent_messages, [])
        # We rollback to the savepoint, so mail 2 should never be send anymore.
        savepoint.rollback()
        self.assertEqual(mailer.sent_messages, [])
        # Any mail after this *should* be sent.
        delivery.send(fromaddr, toaddrs, b'Subject: three')
        self.assertEqual(mailer.sent_messages, [])
        transaction.commit()
        self.assertEqual(len(mailer.sent_messages), 2)
        # They might not necessarily be sent in the order we expect.
        # Get the subject lines.
        all_text = b'\n'.join([mail[2] for mail in mailer.sent_messages])
        lines = all_text.splitlines()
        subjects = [line for line in lines if b'Subject' in line]
        subjects.sort()
        self.assertEqual([b'Subject: one', b'Subject: three'], subjects)


class MaildirWriterStub(object):

    data = b''
    commited_messages = []  # this list is shared among all instances
    aborted_messages = []   # this one too
    _closed = False
    _commited = False
    _aborted = False

    def write(self, data):
        if self._closed:
            raise AssertionError('already closed')
        self.data += data

    def writelines(self, seq):
        raise NotImplementedError()

    def close(self):
        self._closed = True

    def commit(self):
        if not self._closed:
            raise AssertionError('for this test we want the message explicitly'
                                 ' closed before it is committed')
        self._commited = True
        self.commited_messages.append(self.data)

    def abort(self):
        if not self._closed:
            raise AssertionError('for this test we want the message explicitly'
                                 ' closed before it is committed')
        self._aborted = True
        self.aborted_messages.append(self.data)


class MaildirStub(object):

    def __init__(self, path, create=False):
        self.path = path
        self.create = create
        self.msgs = []
        self.files = []

    def __iter__(self):
        return iter(self.files)

    def newMessage(self):
        m = MaildirWriterStub()
        self.msgs.append(m)
        return m


class WritableMaildirStub(MaildirStub):

    STUB_DEFAULT_MESSAGE_LINES = (
        b'X-Zope-From: foo@example.com\n',
        b'X-Zope-To: bar@example.com, baz@example.com\n',
        b'Header: value\n\nBody\n')

    # The result of sending a message written with the default lines
    # through the stub mailer
    STUB_DEFAULT_MESSAGE_SENT = (
        'foo@example.com',
        ('bar@example.com', 'baz@example.com'),
        b'Header: value\n\nBody\n')

    STUB_DEFAULT_MESSAGE_RECPT = ('bar@example.com', 'baz@example.com')

    def __init__(self, test, *args, **kwargs):
        super(WritableMaildirStub, self).__init__(*args, **kwargs)
        self.stub_directory = tempfile.mkdtemp(suffix=".test_maildir")
        test.addCleanup(shutil.rmtree, self.stub_directory)

    def stub_createFile(self, filename="message",
                        lines=STUB_DEFAULT_MESSAGE_LINES):
        """
        Create a new file in the temporary directory.

        The filename is just the base portion.
        """
        filename = os.path.join(self.stub_directory, filename)
        with open(filename, 'wb') as f:
            for line in lines:
                f.write(line)
        self.files.append(filename)
        return filename

    def stub_getTmpFilename(self, filename="message"):
        filename = os.path.join(self.stub_directory, filename)
        head, tail = os.path.split(filename)
        tmp_filename = os.path.join(head, '.sending-' + tail)
        return tmp_filename

    def stub_getFailedFilename(self, filename="message"):
        filename = os.path.join(self.stub_directory, filename)
        head, tail = os.path.split(filename)
        fail_filename = os.path.join(head, '.rejected-' + tail)
        return fail_filename

    def stub_createTmpFile(self, filename="message"):
        """
        Create a temporary version of a file that already exists by
        copying its data to a properly named file.

        Returns the complete path.
        """
        tmp_filename = self.stub_getTmpFilename(filename)
        filename = os.path.join(self.stub_directory, filename)
        shutil.copyfile(filename, tmp_filename)
        return tmp_filename


class LoggerStub(object):

    def __init__(self):
        self.infos = []
        self.errors = []

    def getLogger(self, name):
        raise NotImplementedError()

    def error(self, msg, *args, **kwargs):
        error = (msg, args, kwargs)
        if kwargs.get("exc_info"):
            error += sys.exc_info()[:2]
        self.errors.append(error)

    def info(self, msg, *args, **kwargs):
        self.infos.append((msg, args, kwargs))


class BizzarreMailError(IOError):
    pass


@implementer(IMailer)
class BrokenMailerStub(object):

    def __init__(self, *args, **kw):
        pass

    def send(self, fromaddr, toaddrs, message):
        raise BizzarreMailError("bad things happened while sending mail")

    vote = None
    abort = None


@implementer(IMailer)
class RefusingMailerStub(object):

    def __init__(self, *args, **kw):
        pass

    def vote(self, fromaddr, toaddrs, message):
        raise BizzarreMailError("bad things happened while sending mail")

    def send(self, fromaddr, toaddrs, message):
        raise NotImplementedError()

    abort = None


@implementer(IMailer)
class SMTPResponseExceptionMailerStub(object):

    def __init__(self, code):
        self.code = code

    def send(self, fromaddr, toaddrs, message):
        raise smtplib.SMTPResponseException(self.code, 'Serious Error')


class TestQueuedMailDelivery(unittest.TestCase):

    def setUp(self):
        import zope.sendmail.delivery as mail_delivery_module
        self.mail_delivery_module = mail_delivery_module
        self.old_Maildir = mail_delivery_module.Maildir
        mail_delivery_module.Maildir = MaildirStub

    def tearDown(self):
        self.mail_delivery_module.Maildir = self.old_Maildir
        MaildirWriterStub.commited_messages = []
        MaildirWriterStub.aborted_messages = []

    def testInterface(self):
        from zope.sendmail.interfaces import IQueuedMailDelivery
        from zope.sendmail.delivery import QueuedMailDelivery
        delivery = QueuedMailDelivery('/path/to/mailbox')
        verifyObject(IQueuedMailDelivery, delivery)
        self.assertEqual(delivery.queuePath, '/path/to/mailbox')

    def testSend(self):
        from zope.sendmail.delivery import QueuedMailDelivery
        delivery = QueuedMailDelivery('/path/to/mailbox')
        fromaddr = 'jim@example.com'
        toaddrs = ('guido@example.com',
                   'steve@examplecom')
        zope_headers = (b'X-Zope-From: jim@example.com\n'
                        b'X-Zope-To: guido@example.com, steve@examplecom\n')
        opt_headers = (b'From: Jim <jim@example.org>\n'
                       b'To: some-zope-coders:;\n'
                       b'Date: Mon, 19 May 2003 10:17:36 -0400\n'
                       b'Message-Id: <20030519.1234@example.org>\n')
        message = (b'Subject: example\n'
                   b'\n'
                   b'This is just an example\n')

        msgid = delivery.send(fromaddr, toaddrs, opt_headers + message)
        self.assertEqual(msgid, '20030519.1234@example.org')
        self.assertEqual(MaildirWriterStub.commited_messages, [])
        self.assertEqual(MaildirWriterStub.aborted_messages, [])
        transaction.commit()
        self.assertEqual(MaildirWriterStub.commited_messages,
                         [zope_headers + opt_headers + message])
        self.assertEqual(MaildirWriterStub.aborted_messages, [])

        MaildirWriterStub.commited_messages = []
        msgid = delivery.send(fromaddr, toaddrs, message)
        self.assertIn('@', msgid)
        self.assertEqual(MaildirWriterStub.commited_messages, [])
        self.assertEqual(MaildirWriterStub.aborted_messages, [])
        transaction.commit()
        self.assertEqual(len(MaildirWriterStub.commited_messages), 1)
        self.assertTrue(
            MaildirWriterStub.commited_messages[0].endswith(message)
        )
        new_headers = MaildirWriterStub.commited_messages[0][:-len(message)]
        self.assertIn(('Message-Id: <%s>' % msgid).encode(), new_headers)
        self.assertIn(('X-Zope-From: %s' % fromaddr).encode(), new_headers)
        self.assertIn(('X-Zope-To: %s' % ", ".join(toaddrs)).encode(),
                      new_headers)
        self.assertEqual(MaildirWriterStub.aborted_messages, [])

        MaildirWriterStub.commited_messages = []
        msgid = delivery.send(fromaddr, toaddrs, opt_headers + message)
        self.assertEqual(MaildirWriterStub.commited_messages, [])
        self.assertEqual(MaildirWriterStub.aborted_messages, [])
        transaction.abort()
        self.assertEqual(MaildirWriterStub.commited_messages, [])
        self.assertEqual(len(MaildirWriterStub.aborted_messages), 1)



##############################################################################
#
# Copyright (c) 2003 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Mail Delivery utility implementation

This module contains various implementations of `MailDeliverys`.
"""
__docformat__ = 'restructuredtext'

import os
import email.parser
import logging
import warnings
from random import randrange
from time import strftime
from socket import gethostname

from transaction.interfaces import ISavepointDataManager
from transaction.interfaces import IDataManagerSavepoint
import transaction

from zope.interface import implementer
from zope.sendmail.interfaces import IDirectMailDelivery, IQueuedMailDelivery
from zope.sendmail.maildir import Maildir

# BBB: this import is needed for backward compatibility with older versions of
# zope.sendmail which defined QueueProcessorThread in this module
from zope.sendmail.queue import QueueProcessorThread  # noqa: F401
from ._compat import PY2


log = logging.getLogger("MailDataManager")


@implementer(IDataManagerSavepoint)
class _NoOpSavepoint(object):

    def rollback(self):
        return


@implementer(ISavepointDataManager)
class MailDataManager(object):

    def __init__(self, callable, args=(), vote=None, onAbort=None):
        self.callable = callable
        self.args = args
        self.vote = vote
        self.onAbort = onAbort
        # Use the default thread transaction manager.
        self.transaction_manager = transaction.manager

    def commit(self, txn):
        pass

    def abort(self, txn):
        if self.onAbort:
            self.onAbort()

    def sortKey(self):
        return str(id(self))

    def savepoint(self):
        # We do not need savepoint/rollback, but some code (like CMFEditions)
        # uses savepoints, and breaks when one datamanager does not have this.
        # So provide a dummy implementation.
        return _NoOpSavepoint()

    # No subtransaction support.
    def abort_sub(self, txn):
        "This object does not do anything with subtransactions"

    commit_sub = abort_sub

    def beforeCompletion(self, txn):
        "This object does not do anything in beforeCompletion"

    afterCompletion = beforeCompletion

    def tpc_begin(self, txn, subtransaction=False):
        assert not subtransaction

    def tpc_vote(self, txn):
        if self.vote is not None:
            return self.vote(*self.args)

    def tpc_finish(self, txn):
        try:
            self.callable(*self.args)
        except Exception:
            # Any exceptions here can cause database corruption.
            # Better to protect the data and potentially miss emails than
            # leave a database in an inconsistent state which requires a
            # guru to fix.
            log.exception("Failed in tpc_finish for %r", self.callable)

    tpc_abort = abort


class AbstractMailDelivery(object):

    def newMessageId(self):
        """Generates a new message ID according to RFC 2822 rules"""
        randmax = 0x7fffffff
        left_part = '%s.%d.%d' % (strftime('%Y%m%d%H%M%S'),
                                  os.getpid(),
                                  randrange(0, randmax))
        return "%s@%s" % (left_part, gethostname())

    def send(self, fromaddr, toaddrs, message):
        # Switch the message to be bytes immediately, any encoding
        # peculiarities should be handled before.
        if message is None:
            header = b''
        else:
            if not isinstance(message, bytes):
                message = message.encode('utf-8')
            # determine line separator type (assumes consistency)
            nli = message.find(b'\n')
            line_sep = b'\n' if nli < 1 or message[nli - 1] != b'\r' \
                else b'\r\n'
            header = message.split(line_sep * 2, 1)[0]

        if PY2:
            parse = email.parser.Parser().parsestr  # pragma: PY2
        else:
            parse = email.parser.BytesParser().parsebytes  # pragma: PY3
        messageid = parse(header).get('Message-Id')
        if messageid:
            if not messageid.startswith('<') or not messageid.endswith('>'):
                raise ValueError('Malformed Message-Id header')
            messageid = messageid[1:-1]
        else:
            messageid = self.newMessageId()
            message = b'Message-Id: <%s>\n%s' % (messageid.encode(), message)
        transaction.get().join(
            self.createDataManager(fromaddr, toaddrs, message))
        return messageid

    def createDataManager(self, fromaddr, toaddrs, message):
        raise NotImplementedError()


@implementer(IDirectMailDelivery)
class DirectMailDelivery(AbstractMailDelivery):
    __doc__ = IDirectMailDelivery.__doc__

    def __init__(self, mailer):
        self.mailer = mailer

    def createDataManager(self, fromaddr, toaddrs, message):
        try:
            vote = self.mailer.vote
        except AttributeError:
            # We've got an old mailer, just pass through voting
            warnings.warn("The mailer %s does not provide a vote method"
                          % (repr(self.mailer)), DeprecationWarning)

            def vote(*args, **kwargs):
                pass

        return MailDataManager(self.mailer.send,
                               args=(fromaddr, toaddrs, message),
                               vote=vote,
                               onAbort=self.mailer.abort)


@implementer(IQueuedMailDelivery)
class QueuedMailDelivery(AbstractMailDelivery):
    __doc__ = IQueuedMailDelivery.__doc__

    def __init__(self, queuePath):
        self._queuePath = queuePath

    queuePath = property(lambda self: self._queuePath)

    def createDataManager(self, fromaddr, toaddrs, message):
        maildir = Maildir(self.queuePath, True)
        msg = maildir.newMessage()
        msg.write(b'X-Zope-From: %s\n' % fromaddr.encode())
        msg.write(b'X-Zope-To: %s\n' % ", ".join(toaddrs).encode())
        msg.write(message)
        msg.close()
        return MailDataManager(msg.commit, onAbort=msg.abort)



##############################################################################
#
# Copyright (c) 2003 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Mail Delivery Tests

Simple implementation of the MailDelivery, Mailers and MailEvents.
"""
import errno
import os.path
import shutil
import sys
from tempfile import mkdtemp
import unittest
import io
from contextlib import contextmanager

from zope.sendmail.queue import ConsoleApp
from zope.sendmail.tests.test_delivery import WritableMaildirStub
from zope.sendmail.tests.test_delivery import LoggerStub
from zope.sendmail.tests.test_delivery import BrokenMailerStub
from zope.sendmail.tests.test_delivery import SMTPResponseExceptionMailerStub
from zope.sendmail.tests.test_delivery import MailerStub
from zope.sendmail.tests.test_delivery import BizzarreMailError

from zope.sendmail import queue


class QPTesting(queue.QueueProcessorThread):

    test = None

    def _makeMaildir(self, path):
        return WritableMaildirStub(self.test, path)


class SMTPRecipientsRefusedMailerStub(object):

    def __init__(self, recipients):
        self.recipients = recipients

    def send(self, fromaddr, toaddrs, message):
        import smtplib
        raise smtplib.SMTPRecipientsRefused(self.recipients)


@contextmanager
def patched(module, attr, func):
    orig = getattr(module, attr)
    setattr(module, attr, func)
    try:
        yield
    finally:
        setattr(module, attr, orig)


class TestQueueProcessorThread(unittest.TestCase):

    def setUp(self):
        self.thread = QPTesting()
        self.thread.test = self
        self.thread.setQueuePath('/foo/bar/baz')
        self.md = self.thread.maildir
        self.dir = self.md.stub_directory
        self.mailer = MailerStub()
        self.thread.setMailer(self.mailer)
        self.thread.log = LoggerStub()
        self.filename = None

    def _assertEmptyErrorLog(self):
        self.assertEqual(self.thread.log.errors, [])

    def _assertErrorLog(
        self,
        From=WritableMaildirStub.STUB_DEFAULT_MESSAGE_SENT[0],
        to=", ".join(WritableMaildirStub.STUB_DEFAULT_MESSAGE_SENT[1]),
        exception_kind=None,
    ):
        expected = ('Error while sending mail from %s to %s.',
                    (From, to,),
                    {'exc_info': True},)
        error = self.thread.log.errors[0]
        if exception_kind:
            error = error[:4]
            expected += (exception_kind,)
        else:
            error = error[:3]
        self.assertEqual(error, expected)

    def _assertGenericErrorLog(self, filename="message",
                               exception=None):
        full_path = os.path.join(self.dir, filename)
        expected = ('Error while sending mail : %s ',
                    (full_path,),
                    {'exc_info': True})
        error = self.thread.log.errors[0]
        error = error[:5]
        expected += (type(exception), exception,)
        self.assertEqual(error, expected)

    def _assertTmpMessagePathDoesNotExist(self, filename="message"):
        full_path = self.md.stub_getTmpFilename(filename)
        self.assertFalse(
            os.path.exists(full_path),
            "The temporary path '%s' should not exist" % full_path)

    def _assertTmpMessagePathExists(self, filename="message"):
        full_path = self.md.stub_getTmpFilename(filename)
        self.assertTrue(os.path.exists(full_path),
                        "The temporary path '%s' should exist" % full_path)

    def _assertRejectedMessagePathExists(self, filename="message"):
        full_path = self.md.stub_getFailedFilename(filename)
        self.assertTrue(os.path.exists(full_path),
                        "The rejected path '%s' should exist" % full_path)

    def _assertMessagePathExists(self, filename="message"):
        full_path = os.path.join(self.dir, filename)
        self.assertTrue(os.path.exists(full_path),
                        "The path '%s' should exist" % full_path)

    def _assertMessagePathDoesNotExist(self, filename="message"):
        full_path = os.path.join(self.dir, filename)
        self.assertFalse(os.path.exists(full_path),
                         "The path '%s' should not exist" % full_path)

    def test_makeMaildir_creates(self):
        md = queue.QueueProcessorThread()._makeMaildir(
            os.path.join(self.dir, 'testing'))
        self.assertTrue(os.path.exists(md.path))

    def test_threadName(self):
        self.assertEqual(self.thread.name,
                         "zope.sendmail.queue.QueueProcessorThread")

    def test_parseMessage(self):
        hdr = (b'X-Zope-From: foo@example.com\n'
               b'X-Zope-To: bar@example.com, baz@example.com\n')
        msg = (b'Header: value\n'
               b'\n'
               b'Body\n')
        f, t, m = self.thread._parseMessage(hdr + msg)
        self.assertEqual(f, 'foo@example.com')
        self.assertEqual(t, ('bar@example.com', 'baz@example.com'))
        self.assertEqual(m, msg)

    def test_parseMessage_empty_address(self):
        hdr = (b'X-Zope-From: foo@example.com\n'
               b'X-Zope-To: bar@example.com, , baz@example.com\n')
        msg = (b'Header: value\n'
               b'\n'
               b'Body\n')
        f, t, m = self.thread._parseMessage(hdr + msg)
        self.assertEqual(f, 'foo@example.com')
        self.assertEqual(t, ('bar@example.com', 'baz@example.com'))
        self.assertEqual(m, msg)

    def test_parseMessage_error(self):
        msg = b"bad message"
        f, t, m = self.thread._parseMessage(msg)
        self.assertEqual(f, "")
        self.assertEqual(t, ())
        self.assertEqual(m, msg)

    def test_deliveration(self):
        self.filename = self.md.stub_createFile('message')
        self._assertMessagePathExists("message")
        self.thread.run(forever=False)
        self.assertEqual(self.mailer.sent_messages,
                         [self.md.STUB_DEFAULT_MESSAGE_SENT])
        self._assertMessagePathDoesNotExist('message')
        self.assertEqual(self.thread.log.infos,
                         [('Mail from %s to %s sent.',
                           ('foo@example.com',
                            'bar@example.com, baz@example.com'),
                           {})])

    def test_error_logging(self):
        self.thread.setMailer(BrokenMailerStub())
        self.filename = self.md.stub_createFile('message')
        self._assertMessagePathExists("message")
        self.thread.run(forever=False)
        self._assertErrorLog(exception_kind=BizzarreMailError)

    def test_smtp_response_error_transient(self):
        # Test a transient error
        self.thread.setMailer(SMTPResponseExceptionMailerStub(451))
        self.filename = self.md.stub_createFile('message')

        self.thread.run(forever=False)

        # File must remain were it was, so it will be retried
        self._assertMessagePathExists("message")
        self._assertErrorLog()

    def test_smtp_response_error_permanent(self):
        # Test a permanent error
        self.thread.setMailer(SMTPResponseExceptionMailerStub(550))
        self.filename = self.md.stub_createFile('message')
        self._assertMessagePathExists("message")

        self.thread.run(forever=False)

        # File must be moved aside
        self._assertMessagePathDoesNotExist("message")
        self._assertRejectedMessagePathExists('message')

        self.assertEqual(self.thread.log.errors,
                         [('Discarding email from %s to %s due to a '
                           'permanent error: %s',
                           ('foo@example.com',
                            'bar@example.com, baz@example.com',
                            "(550, 'Serious Error')"), {})])

    def test_smtp_recipients_refused(self):
        # Test a permanent error
        self.thread.setMailer(SMTPRecipientsRefusedMailerStub(
            [self.md.STUB_DEFAULT_MESSAGE_RECPT[0]]))
        self.md.stub_createFile()
        self.thread.run(forever=False)

        # File must be moved aside
        self._assertMessagePathDoesNotExist()
        self._assertRejectedMessagePathExists()

        self.assertEqual(self.thread.log.errors,
                         [('Email recipients refused: %s',
                           (self.md.STUB_DEFAULT_MESSAGE_RECPT[0],), {})])

    def test_stop_while_running(self):
        test = self

        class Maildir(object):
            count = 0

            def __iter__(self):
                return self

            def __next__(self):
                self.count += 1
                if self.count == 1:
                    test.thread.stop()
                    return
                raise AssertionError("Should have stopped")

            next = __next__  # Python 2

        self.thread.setMaildir(Maildir())
        self.thread.run()
        self.assertEqual(1, self.thread.maildir.count)

    def test_tmpfile_exists(self):
        self.md.stub_createFile()
        self.md.stub_createTmpFile()

        self.thread.run(forever=False)

        # Both of them are still there because this was a brand new
        # file
        self._assertMessagePathExists()
        self._assertTmpMessagePathExists()

    def test_tmpfile_cannot_stat(self):
        self.md.stub_createFile()
        tmp_file = self.md.stub_createTmpFile()

        err = OSError(tmp_file)

        def stat(fname):
            # Note that this interferes with debuggers
            self.assertEqual(fname, tmp_file)
            raise err

        with patched(os, 'stat', stat):
            self.thread.run(forever=False)

        # Both of them are still there because of the exception
        self._assertMessagePathExists()
        self._assertTmpMessagePathExists()

        # And we logged the random error
        self._assertGenericErrorLog(exception=err)

    def test_tmpfile_too_old(self):
        self.md.stub_createFile()
        self.md.stub_createTmpFile()

        with patched(queue, 'MAX_SEND_TIME', -1):
            self.thread.run(forever=False)

        # Neither of them are there any more
        # file
        self._assertMessagePathDoesNotExist()
        self._assertTmpMessagePathDoesNotExist()

    def test_tmpfile_too_old_unlink_fails_generic(self):
        # generic unlink exceptions are ignored and processing just continues.
        # Later we fail to link the message file to the tempfile and catch that
        # and bail.
        self.md.stub_createFile()
        tmp_file = self.md.stub_createTmpFile()

        def unlink(fname):
            self.assertEqual(fname, tmp_file)
            raise OSError(fname)

        with patched(queue, 'MAX_SEND_TIME', -1):
            with patched(os, 'unlink', unlink):
                self.thread.run(forever=False)

        # The message was not processed
        self._assertMessagePathExists()
        # And our patch kept the tmpfile in place
        self._assertTmpMessagePathExists()
        self._assertEmptyErrorLog()

    def test_tmpfile_too_old_unlink_fails_dne(self):
        # If we try to remove a tempfile and its already gone,
        # skip it
        self.md.stub_createFile()
        tmp_file = self.md.stub_createTmpFile()

        def unlink(fname):
            self.assertEqual(fname, tmp_file)
            raise OSError(errno.ENOENT, fname)

        with patched(queue, 'MAX_SEND_TIME', -1):
            with patched(os, 'unlink', unlink):
                self.thread.run(forever=False)

        # The message was not processed
        self._assertMessagePathExists()
        # But our patch kept the tmpfile in place
        self._assertTmpMessagePathExists()
        self._assertEmptyErrorLog()

    def test_utime_fails_dne(self):
        # A DNE error from utime causes the process to continue
        filename = self.md.stub_createFile()

        def utime(fname, atm):
            self.assertEqual(fname, filename)
            self.assertIsNone(atm)
            raise OSError(errno.ENOENT, fname)

        with patched(os, 'utime', utime):
            self.thread.run(forever=False)

        # The message was not processed
        self._assertMessagePathExists()
        self._assertEmptyErrorLog()

    def test_run_forever(self):
        import time

        class DoneSleeping(Exception):
            pass

        def sleep(i):
            self.assertEqual(i, self.thread.interval)
            raise DoneSleeping()

        with patched(time, 'sleep', sleep):
            with self.assertRaises(DoneSleeping):
                self.thread.run()


test_ini = """[app:zope-sendmail]
interval = 33
hostname = testhost
port = 2525
username = Chris
password = Rossi
force_tls = False
no_tls = True
queue_path = hammer/dont/hurt/em
"""


class TestConsoleApp(unittest.TestCase):
    def setUp(self):
        from zope.sendmail.delivery import QueuedMailDelivery
        from zope.sendmail.maildir import Maildir
        self.dir = mkdtemp()
        self.queue_dir = os.path.join(self.dir, "queue")
        self.delivery = QueuedMailDelivery(self.queue_dir)
        self.maildir = Maildir(self.queue_dir, True)
        self.mailer = MailerStub()
        io_type = io.StringIO if bytes is not str else io.BytesIO
        self.stderr = io_type()
        self.stdout = io_type()

    def tearDown(self):
        shutil.rmtree(self.dir)

    def _make_one(self, cmdline):
        cmdline = cmdline.split() if isinstance(cmdline, str) else cmdline
        with patched(sys, 'stdout', self.stdout), \
                patched(sys, 'stderr', self.stderr):
            return ConsoleApp(cmdline, verbose=False)

    def _get_output(self):
        return self.stderr.getvalue() + self.stdout.getvalue()

    def test_args_processing(self):
        # simplest case that works
        cmdline = "zope-sendmail %s" % self.dir
        app = self._make_one(cmdline)
        self.assertEqual("zope-sendmail", app.script_name)
        self.assertEqual(self.dir, app.queue_path)
        self.assertFalse(app.daemon)
        self.assertEqual(3, app.interval)
        self.assertEqual("localhost", app.hostname)
        self.assertEqual(25, app.port)
        self.assertEqual(None, app.username)
        self.assertEqual(None, app.password)
        self.assertFalse(app.force_tls)
        self.assertFalse(app.no_tls)

    def test_args_processing_no_queue_path(self):
        # simplest case that doesn't work: no queue path specified
        cmdline = "zope-sendmail"

        with self.assertRaises(SystemExit):
            self._make_one(cmdline)

        self.assertIn('please specify the queue path', self._get_output())

    def test_args_processing_almost_all_options(self):
        # use (almost) all of the options
        cmdline = (
            "zope-sendmail --daemon --interval 7 --hostname foo --port 75 "
            "--username chris --password rossi --force-tls "
            "%s" % self.dir
        )
        app = self._make_one(cmdline)
        self.assertEqual("zope-sendmail", app.script_name)
        self.assertEqual(self.dir, app.queue_path)
        self.assertTrue(app.daemon)
        self.assertEqual(7, app.interval)
        self.assertEqual("foo", app.hostname)
        self.assertEqual(75, app.port)
        self.assertEqual("chris", app.username)
        self.assertEqual("rossi", app.password)
        self.assertTrue(app.force_tls)
        self.assertFalse(app.no_tls)

        # Add an extra argument
        cmdline += ' another-one'
        with self.assertRaises(SystemExit):
            self._make_one(cmdline)

        self.assertIn('unrecognized argument', self._get_output())

    def test_args_processing_username_without_password(self):
        # test username without password
        cmdline = "zope-sendmail --username chris %s" % self.dir

        with self.assertRaises(SystemExit):
            self._make_one(cmdline)

        self.assertIn('Must use username and password together',
                      self._get_output())

    def test_args_processing_force_tls_and_no_tls(self):
        # test force_tls and no_tls
        cmdline = "zope-sendmail --force-tls --no-tls %s" % self.dir

        with self.assertRaises(SystemExit):
            self._make_one(cmdline)

        self.assertIn('--no-tls: not allowed with argument',
                      self._get_output())

    def test_ini_parse(self):
        ini_path = os.path.join(self.dir, "zope-sendmail.ini")
        with open(ini_path, "w") as f:
            f.write(test_ini)
        # override most everything
        cmdline = """zope-sendmail --config %s""" % ini_path
        app = ConsoleApp(cmdline.split(), verbose=False)
        self.assertEqual("zope-sendmail", app.script_name)
        self.assertEqual("hammer/dont/hurt/em", app.queue_path)
        self.assertFalse(app.daemon)
        self.assertEqual(33, app.interval)
        self.assertEqual("testhost", app.hostname)
        self.assertEqual(2525, app.port)
        self.assertEqual("Chris", app.username)
        self.assertEqual("Rossi", app.password)
        self.assertFalse(app.force_tls)
        self.assertTrue(app.no_tls)
        # override nothing, make sure defaults come through
        with open(ini_path, "w") as f:
            f.write("[app:zope-sendmail]\n\nqueue_path=foo\n")
        cmdline = """zope-sendmail --config %s %s""" % (ini_path, self.dir)
        app = ConsoleApp(cmdline.split(), verbose=False)
        self.assertEqual("zope-sendmail", app.script_name)
        self.assertEqual(self.dir, app.queue_path)
        self.assertFalse(app.daemon)
        self.assertEqual(3, app.interval)
        self.assertEqual("localhost", app.hostname)
        self.assertEqual(25, app.port)
        self.assertEqual(None, app.username)
        self.assertEqual(None, app.password)
        self.assertFalse(app.force_tls)
        self.assertFalse(app.no_tls)

    def test_ini_not_exist(self):
        cmdline = ['sendmail', '--config', 'this path cannot be opened']
        with self.assertRaises(SystemExit) as exc:
            self._make_one(cmdline)

        self.assertIn('No such file or directory', self._get_output())
        self.assertEqual(exc.exception.code, 2)

    def test_run(self):
        cmdline = ['sendmail', self.dir]

        with patched(QPTesting, 'test', self), \
                patched(ConsoleApp, 'QueueProcessorKind', QPTesting), \
                patched(ConsoleApp, 'MailerKind', MailerStub):
            queue.run(cmdline)

    def test_help(self):
        cmdline = ['prog', '--help']

        with self.assertRaises(SystemExit) as exc:
            self._make_one(cmdline)

        self.assertIn('usage', self._get_output())
        self.assertEqual(exc.exception.code, 0)



##############################################################################
#
# Copyright (c) 2006 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
# This package is developed by the Zope Toolkit project, documented here:
# http://docs.zope.org/zopetoolkit
# When developing and releasing this package, please follow the documented
# Zope Toolkit policies as described by this documentation.
##############################################################################
"""Setup for zope.sendmail package"""
import os
from setuptools import setup, find_packages

TESTS_REQUIRE = [
    'zope.security',
    'zope.testing',
    'zope.testrunner',
]

EXTRAS_REQUIRE = {
    'test': TESTS_REQUIRE,
    ':sys_platform == "win32"': ['pywin32'],
    'docs': [
        'Sphinx',
        'repoze.sphinx.autointerface',
        'sphinxcontrib-programoutput',
    ]
}


def read(*rnames):
    with open(os.path.join(os.path.dirname(__file__), *rnames)) as f:
        return f.read()


LONG_DESCRIPTION = (
    read('README.rst')
    + '\n' +
    read('CHANGES.rst')
)

setup(
    name='zope.sendmail',
    version='5.3',
    url='https://github.com/zopefoundation/zope.sendmail',
    license='ZPL 2.1',
    description='Zope sendmail',
    author='Zope Foundation and Contributors',
    author_email='zope-dev@zope.org',
    long_description=LONG_DESCRIPTION,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Zope Public License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Operating System :: OS Independent',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Communications :: Email',
        'Framework :: Zope :: 3',
    ],
    packages=find_packages('src'),
    package_dir={'': 'src'},
    namespace_packages=['zope'],
    python_requires='>=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*',
    extras_require=EXTRAS_REQUIRE,
    install_requires=[
        'setuptools',
        'transaction',
        'zope.i18nmessageid',
        'zope.interface',
        'zope.schema',
        # it's only needed for vocabulary, zcml and tests
        'zope.component>=3.8.0',
        # these are only needed for zcml
        'zope.configuration',
    ],
    tests_require=TESTS_REQUIRE,
    test_suite='zope.sendmail.tests',
    include_package_data=True,
    zip_safe=False,
    entry_points="""
      [console_scripts]
      zope-sendmail = zope.sendmail.queue:run
    """
)



# coding=utf-8
##############################################################################
#
# Copyright (c) 2003 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Tests for mailers.
"""

import unittest
from functools import partial
from ssl import SSLError

from zope.interface.verify import verifyObject
from zope.sendmail.interfaces import ISMTPMailer
from zope.sendmail.mailer import SMTPMailer


# This works for both, Python 2 and 3.
port_types = (int, str)


class SMTP(object):

    fail_on_quit = False

    def __init__(self, h, p):
        self.hostname = h
        self.port = p
        self.quitted = False
        self.closed = False
        assert isinstance(p, port_types)

    def sendmail(self, f, t, m):
        assert not self.closed
        assert not self.quitted
        self.fromaddr = f
        self.toaddrs = t
        self.msgtext = m

    def login(self, username, password):
        self.username = username
        self.password = password

    def quit(self):
        if self.fail_on_quit:
            raise SSLError("dang")
        self.quitted = True
        self.close()

    def close(self):
        self.closed = True

    def has_extn(self, ext):
        return True

    ehlo_code = 200
    ehlo_msg = 'Hello, I am your stupid MTA mock'

    def ehlo(self):
        self.does_esmtp = True
        return (self.ehlo_code, self.ehlo_msg)

    def starttls(self):
        pass


class SMTPWithNoEHLO(SMTP):
    does_esmtp = False

    def helo(self):
        return (200, 'Hello, I am your stupid MTA mock')

    ehlo_code = 502
    ehlo_msg = "I don't understand EHL"


class TestSMTPMailer(unittest.TestCase):

    SMTPClass = SMTP

    # Avoid DeprecationWarning for assertRaisesRegexp on Python 3 while
    # coping with Python 2 not having the Regex spelling variant
    assertRaisesRegex = getattr(unittest.TestCase, 'assertRaisesRegex',
                                unittest.TestCase.assertRaisesRegexp)

    def _makeMailer(self, port=None, smtp_hook=None):
        if port is None:
            mailer = SMTPMailer()
        else:
            mailer = SMTPMailer(u'localhost', port)

        def _make_smtp(host, port):
            smtp = self.SMTPClass(host, port)
            if smtp_hook:
                smtp_hook(smtp)
            return smtp

        mailer.smtp = _make_smtp
        return mailer

    def setUp(self, port=None):
        self.smtp = None
        self.mailer = self._makeMailer(
            port=port, smtp_hook=partial(setattr, self, 'smtp'))

    def test_interface(self):
        verifyObject(ISMTPMailer, self.mailer)

    def test_send(self):
        for run in (1, 2):
            if run == 2:
                self.setUp(u'25')
            fromaddr = 'me@example.com'
            toaddrs = ('you@example.com', 'him@example.com')
            msgtext = 'Headers: headers\n\nbodybodybody\n-- \nsig\n'
            self.mailer.send(fromaddr, toaddrs, msgtext)
            self.assertEqual(self.smtp.fromaddr, fromaddr)
            self.assertEqual(self.smtp.toaddrs, toaddrs)
            self.assertEqual(self.smtp.msgtext, msgtext)
            self.assertTrue(self.smtp.quitted)
            self.assertTrue(self.smtp.closed)

    def test_send_multiple_same_mailer(self):
        # The mailer re-opens itself as needed when sending
        # multiple mails.
        smtps = []
        mailer = self._makeMailer(smtp_hook=smtps.append)

        # Note: this test needs to be thread-safe so it must avoid mutating
        # attributes on `self`!

        for run in (1, 2):
            fromaddr = 'me@example.com' + str(run)
            toaddrs = ('you@example.com', 'him@example.com')
            msgtext = 'Headers: headers\n\nbodybodybody\n-- \nsig\n'
            mailer.send(fromaddr, toaddrs, msgtext)
            smtp = smtps[-1]
            self.assertEqual(smtp.fromaddr, fromaddr)
            self.assertEqual(smtp.toaddrs, toaddrs)
            self.assertEqual(smtp.msgtext, msgtext)
            self.assertTrue(smtp.quitted)
            self.assertTrue(smtp.closed)

        self.assertEqual(2, len(smtps))

    def test_send_multiple_threads(self):
        import threading

        results = []

        def run():
            try:
                self.test_send_multiple_same_mailer()
            except BaseException as e:  # pragma: no cover
                results.append(e)
                raise
            else:
                results.append(True)

        threads = []
        for _ in range(2):
            threads.append(threading.Thread(target=run))
        for t in threads:
            t.start()

        for t in threads:
            t.join()

        self.assertEqual([True for _ in threads], results)

    def test_send_auth(self):
        fromaddr = 'me@example.com'
        toaddrs = ('you@example.com', 'him@example.com')
        msgtext = 'Headers: headers\n\nbodybodybody\n-- \nsig\n'
        self.mailer.username = 'foo'
        self.mailer.password = 'evil'
        self.mailer.hostname = 'spamrelay'
        self.mailer.port = 31337
        self.mailer.send(fromaddr, toaddrs, msgtext)
        self.assertEqual(self.smtp.username, 'foo')
        self.assertEqual(self.smtp.password, 'evil')
        self.assertEqual(self.smtp.hostname, 'spamrelay')
        self.assertEqual(self.smtp.port, '31337')
        self.assertEqual(self.smtp.fromaddr, fromaddr)
        self.assertEqual(self.smtp.toaddrs, toaddrs)
        self.assertEqual(self.smtp.msgtext, msgtext)
        self.assertTrue(self.smtp.quitted)
        self.assertTrue(self.smtp.closed)

    def test_send_auth_unicode(self):
        fromaddr = 'me@example.com'
        toaddrs = ('you@example.com', 'him@example.com')
        msgtext = 'Headers: headers\n\nbodybodybody\n-- \nsig\n'
        self.mailer.username = u'f\u00f8\u00f8'  # double o slash
        self.mailer.password = u'\u00e9vil'  # e acute
        self.mailer.hostname = 'spamrelay'
        self.mailer.port = 31337
        self.mailer.send(fromaddr, toaddrs, msgtext)
        self.assertEqual(self.smtp.username, 'føø')
        self.assertEqual(self.smtp.password, 'évil')

    def test_send_auth_nonascii(self):
        fromaddr = 'me@example.com'
        toaddrs = ('you@example.com', 'him@example.com')
        msgtext = 'Headers: headers\n\nbodybodybody\n-- \nsig\n'
        self.mailer.username = b'f\xc3\xb8\xc3\xb8'  # double o slash
        self.mailer.password = b'\xc3\xa9vil'  # e acute
        self.mailer.hostname = 'spamrelay'
        self.mailer.port = 31337
        self.mailer.send(fromaddr, toaddrs, msgtext)
        self.assertEqual(self.smtp.username, b'f\xc3\xb8\xc3\xb8')
        self.assertEqual(self.smtp.password, b'\xc3\xa9vil')

    def test_send_failQuit(self):
        SMTP.fail_on_quit = True
        self.addCleanup(lambda: setattr(SMTP, 'fail_on_quit', False))

        fromaddr = 'me@example.com'
        toaddrs = ('you@example.com', 'him@example.com')
        msgtext = 'Headers: headers\n\nbodybodybody\n-- \nsig\n'
        self.mailer.send(fromaddr, toaddrs, msgtext)
        self.assertEqual(self.smtp.fromaddr, fromaddr)
        self.assertEqual(self.smtp.toaddrs, toaddrs)
        self.assertEqual(self.smtp.msgtext, msgtext)
        self.assertTrue(not self.smtp.quitted)
        self.assertTrue(self.smtp.closed)

    def test_vote_bad_connection(self):

        def hook(smtp):
            smtp.ehlo_code = 100
            smtp.helo = lambda: (100, "Nope")

        mailer = self._makeMailer(smtp_hook=hook)

        with self.assertRaisesRegex(RuntimeError,
                                    "Error sending HELO to the SMTP server"):
            mailer.vote(None, None, None)

    def test_abort_no_conn(self):
        self.assertIsNone(self.mailer.abort())

    def test_abort_fails_call_close(self):
        class Conn(object):
            closed = False

            def quit(self):
                raise SSLError()

            def close(self):
                self.closed = True

        conn = Conn()
        self.mailer.connection = conn
        self.mailer.abort()

        self.assertTrue(conn.closed)

    def test_send_no_tls_forced(self):
        class Conn(object):
            def has_extn(self, name):
                assert name == 'starttls'
                return False

        self.mailer.force_tls = True
        self.mailer.connection = Conn()

        with self.assertRaisesRegex(RuntimeError,
                                    'TLS is not available'):
            self.mailer.send(None, None, None)

    def test_send_no_esmtp_with_username(self):
        class Conn(object):
            does_esmtp = False

            def has_extn(self, *args):
                return False

        self.mailer.connection = Conn()
        self.mailer.username = 'user'
        with self.assertRaisesRegex(
                RuntimeError,
                "Mailhost does not support ESMTP but a username"):
            self.mailer.send(None, None, None)


class TestSMTPMailerWithNoEHLO(TestSMTPMailer):

    SMTPClass = SMTPWithNoEHLO

    def test_send_auth(self):
        self.skipTest(
            "This test requires ESMTP, which we're intentionally not enabling")

    test_send_auth_unicode = test_send_auth
    test_send_auth_nonascii = test_send_auth
